'''
synbiochem (c) University of Manchester 2015

synbiochem is licensed under the MIT License.

To view a copy of this license, visit <http://opensource.org/licenses/MIT/>.

@author:  neilswainston
'''
# pylint: disable=too-many-public-methods
import unittest

import synbiochem.utils.math_utils as math_utils


class Test(unittest.TestCase):
    '''Test class for math_utils.'''

    def test_isclose(self):
        '''Tests isclose method.'''
        self.assertTrue(math_utils.isclose(18.015054684, 18.015054684))
        self.assertFalse(math_utils.isclose(18.015054684, 18.015054))

    def test_linprog(self):
        '''Tests linprog method.'''
        a_matrix = [[-238, -142, 1, 248, 131, -1, -2, 1, 2],
                    [-144, -86, 0, 150, 80, 0, 0, 0, 0],
                    [-2, 0, 0, 2, 0, 0, 0, 0, 0],
                    [-57, -9, 0, 62, 4, 0, -1, 0, 1],
                    [-2, -1, 0, 2, 1, 0, 0, 0, 0],
                    [2.0, 1.0, 1.0, -2.0, -2.0, -1, 0, 1, 0]]

        bounds = [(1, 8), (1, 8), (1, 8), (1, 8), (1, 8), (0, 8), (0, 8),
                  (0, 8), (0, 8)]

        self.assertTrue(math_utils.linprog([1] * len(a_matrix[0]), a_matrix,
                                           [0] * len(a_matrix), bounds)[0])


if __name__ == "__main__":
    # import sys;sys.argv = ['', 'Test.testName']
    unittest.main()



'''
synbiochem (c) University of Manchester 2015

synbiochem is licensed under the MIT License.

To view a copy of this license, visit <http://opensource.org/licenses/MIT/>.

@author:  neilswainston
'''
import itertools
import re

import synbiochem.utils.math_utils as math_utils


__ELEMENTAL_MASSES = {
    'H': 1.007825032,
    'He': 3.01602932,
    'Li': 6.015122887,
    'Be': 9.012183065,
    'B': 10.01293695,
    'C': 12,
    'N': 14.003074,
    'O': 15.99491462,
    'F': 18.99840316,
    'Ne': 19.99244018,
    'Na': 22.98976928,
    'Mg': 23.9850417,
    'Al': 26.98153853,
    'Si': 27.97692653,
    'P': 30.973762,
    'S': 31.97207117,
    'Cl': 34.96885268,
    'Ar': 35.96754511,
    'K': 38.96370649,
    'Ca': 39.96259086,
    'Sc': 44.95590828,
    'Ti': 45.95262772,
    'Cr': 49.94604183,
    'V': 49.94715601,
    'Fe': 53.93960899,
    'Mn': 54.93804391,
    'Ni': 57.93534241,
    'Co': 58.93319429,
    'Cu': 62.92959772,
    'Zn': 63.92914201,
    'Ga': 68.9255735,
    'Ge': 69.92424875,
    'Se': 73.92247593,
    'As': 74.92159457,
    'Kr': 77.92036494,
    'Br': 78.9183376,
    'Sr': 83.9134191,
    'Rb': 84.91178974,
    'Y': 88.9058403,
    'Zr': 89.9046977,
    'Mo': 91.90680796,
    'Nb': 92.906373,
    'Ru': 95.90759025,
    'Tc': 96.9063667,
    'Pd': 101.9056022,
    'Rh': 102.905498,
    'Cd': 105.9064599,
    'Ag': 106.9050916,
    'Sn': 111.9048239,
    'In': 112.9040618,
    'Te': 119.9040593,
    'Sb': 120.903812,
    'Xe': 123.905892,
    'I': 126.9044719,
    'Ba': 129.9063207,
    'Cs': 132.905452,
    'Ce': 135.9071292,
    'La': 137.9071149,
    'Pr': 140.9076576,
    'Nd': 141.907729,
    'Sm': 143.9120065,
    'Pm': 144.9127559,
    'Eu': 150.9198578,
    'Gd': 151.9197995,
    'Dy': 155.9242847,
    'Tb': 158.9253547,
    'Er': 161.9287884,
    'Ho': 164.9303288,
    'Yb': 167.9338896,
    'Tm': 168.9342179,
    'Hf': 173.9400461,
    'Lu': 174.9407752,
    'W': 179.9467108,
    'Ta': 179.9474648,
    'Os': 183.9524885,
    'Re': 184.9529545,
    'Pt': 189.9599297,
    'Ir': 190.9605893,
    'Hg': 195.9658326,
    'Au': 196.9665688,
    'Tl': 202.9723446,
    'Pb': 203.973044,
    'Bi': 208.9803991,
    'Po': 208.9824308,
    'At': 209.9871479,
    'Rn': 210.9906011,
    'Ra': 223.0185023,
    'Fr': 223.019736,
    'Ac': 227.0277523,
    'Th': 230.0331341,
    'Pa': 231.0358842,
    'U': 233.0396355,
    'Np': 236.04657,
    'Pu': 238.0495601,
    'Am': 241.0568293,
    'Cm': 243.0613893,
    'Bk': 247.0703073,
    'Cf': 249.0748539,
    'Es': 252.08298,
    'Fm': 257.0951061,
    'Md': 258.0984315,
    'No': 259.10103,
    'Lr': 262.10961,
    'Rf': 267.12179,
    'Db': 268.12567,
    'Hs': 270.13429,
    'Sg': 271.13393,
    'Bh': 272.13826,
    'Mt': 276.15159,
    'Rg': 280.16514,
    'Ds': 281.16451,
    'Uut': 284.17873,
    'Cn': 285.17712,
    'Uup': 288.19274,
    'Fl': 289.19042,
    'Uus': 292.20746,
    'Lv': 293.20449,
    'Uuo': 294.21392,
}


def get_molecular_mass(formula):
    '''Calculate and return molecular mass from chemical formula.'''
    return sum([__ELEMENTAL_MASSES[element] * count
                if element in __ELEMENTAL_MASSES
                else float('NaN')
                for element, count in get_elem_comp(formula).iteritems()])


def get_elem_comp(formula):
    '''Gets elemental composition as a dict from formula.'''
    elem_comp = {}

    for term in re.findall('[A-Z]{1}[0-9]*[a-z]{0,1}[0-9]*', formula):
        element = re.search('[A-z]*', term).group(0)
        result = re.search('[0-9]+', term)
        elem_comp[element] = int(result.group(0)) if result else 1

    return elem_comp


def parse_equation(equation, separator='='):
    '''Parses chemical equation strings.'''
    equation_terms = [re.split('\\s+\\+\\s+', equation_side)
                      for equation_side in
                      re.split('\\s*' + separator + '\\s*', equation)]

    # Add reactants and products:
    return _get_reaction_participants(equation_terms[0], -1) + \
        _get_reaction_participants(equation_terms[1], 1)


def balance(reaction_def, optional_comp=None, max_stoich=8.0):
    '''Applies linear programming to balance reaction.'''
    if optional_comp is None:
        optional_comp = [('H', 1, 'CHEBI:24636'), ('H2O', 0, 'CHEBI:15377')]

    all_formulae = [[x[0] for x in reaction_def if x[2] <= 0],
                    [x[0] for x in reaction_def if x[2] > 0]]
    all_formulae.extend([[x[0] for x in optional_comp] for _ in range(2)])

    all_charges = [[x[1] for x in reaction_def if x[2] <= 0],
                   [x[1] for x in reaction_def if x[2] > 0]]
    all_charges.extend([[x[1] for x in optional_comp] for _ in range(2)])

    all_ids = [[x[3] for x in reaction_def if x[2] <= 0],
               [x[3] for x in reaction_def if x[2] > 0]]
    all_ids.extend([[x[2] for x in optional_comp] for _ in range(2)])

    all_elem_comp = [[_get_elem_comp(formula, idx)
                      for formula in formulae]
                     for idx, formulae in enumerate(all_formulae)]

    (balanced, stoichs) = _optimise(_get_elem_matrix(all_elem_comp,
                                                     all_charges),
                                    max_stoich, len(optional_comp))

    balanced_def = sorted(_get_reaction_def(stoichs, all_formulae,
                                            all_charges, all_ids)) \
        if balanced else reaction_def

    return balanced, \
        balanced and _compare_reaction_defs(reaction_def, balanced_def), \
        balanced_def


def _get_reaction_participants(equation_term, stoich_factor):
    '''Adds reaction participants to a list of participants.'''
    if len(equation_term) == 1 and len(equation_term[0]) == 0:
        return []

    all_terms = [participant.split() for participant in equation_term]
    return [[terms[0], stoich_factor]
            if len(terms) == 1
            else [terms[1], stoich_factor * float(terms[0])]
            for terms in all_terms]


def _get_elem_comp(formula, idx):
    '''Returns elemental compositions, multiplying stoichiometry by -1
    if reactants.'''
    elem_comp = get_elem_comp(formula)
    elem_comp.update((x, y * (-1 if idx % 2 == 0 else 1))
                     for x, y in elem_comp.items())
    return elem_comp


def _get_elem_matrix(all_elem_comp, all_charges):
    '''Gets the elemental (and charge) matrix representing rows of elements
    and columns of compounds.'''
    a_matrix = []
    elements = [elem_comp.keys() for elem_comps in all_elem_comp
                for elem_comp in elem_comps]

    for element in set([item for sublist in elements for item in sublist]):
        a_matrix.append([elem_comp[element] if element in elem_comp else 0
                         for elem_comps in all_elem_comp
                         for elem_comp in elem_comps])

    a_matrix.append(list(itertools.chain(*([charge * (-1 if idx % 2 == 0
                                                      else 1)
                                            for charge in charges]
                                           for idx, charges
                                           in enumerate(all_charges)))))

    return a_matrix


def _optimise(a_matrix, max_stoich, num_opt_formulae):
    '''Optimise linear program and return result.'''
    bounds = [(1.0, max_stoich)] * (len(a_matrix[0]) -
                                    num_opt_formulae * 2) + \
        [(0.0, max_stoich)] * num_opt_formulae * 2

    return math_utils.linprog([1] * len(a_matrix[0]),
                              a_matrix,
                              [0.0] * (len(a_matrix)),
                              bounds)


def _get_reaction_def(stoichs, all_formulae, all_charges, all_ids):
    '''Formats the input into (formula, charge, stoichiometry) tuples.'''
    return [(a[0], b, a[1] * c, d)
            for a, b, c, d in zip([(x, -1 if idx % 2 == 0 else 1)
                                   for idx, formulae in enumerate(all_formulae)
                                   for x in formulae],
                                  [y for charges in all_charges
                                   for y in charges],
                                  _simplify_stoichs(stoichs),
                                  [z for ids in all_ids for z in ids])
            if c > 1e-8]


def _compare_reaction_defs(def1, def2):
    '''Returns True/False depending on whether reaction definitions are
    equal.'''
    return def1 is not None and def2 is not None and \
        len(def1) == len(def2) and \
        all([x[0] == y[0] and x[1] == y[1] and math_utils.isclose(x[2], y[2])
             for x, y in zip(sorted(def1), sorted(def2))])


def _simplify_stoichs(stoichs):
    '''Attempts to simplify stoichs of 1.00000001 to 1.0.'''
    return [float(round(stoich)) if math_utils.isclose(stoich, round(stoich))
            else stoich for stoich in stoichs]



'''
synbiochem (c) University of Manchester 2015

synbiochem is licensed under the MIT License.

To view a copy of this license, visit <http://opensource.org/licenses/MIT/>.

@author:  neilswainston
'''
# pylint: disable=no-member
# pylint: disable=too-few-public-methods
# pylint: disable=too-many-arguments
import abc
import random

import numpy


class Chromosome(object):
    '''Class to represent a chromosome.'''

    def __init__(self, chromosome, len_chromosome, mutation_rate=0.01):
        self.__chromosome = chromosome
        self.__len_chromosome = len_chromosome
        self.__mask = 2 ** (self.__len_chromosome) - 1
        self.__mutation_rate = mutation_rate

    def get_chromosome(self):
        '''Gets the chromosome.'''
        return self.__chromosome

    def mutate(self):
        '''Mutates the chromosome.'''
        for i in range(self.__len_chromosome):
            if random.random() < self.__mutation_rate:
                # Mutate:
                mask = 1 << i
                self.__chromosome = self.__chromosome ^ mask

    def breed(self, partner):
        '''Breeds chromosome with a partner Chromosome.'''
        i = int(random.random() * self.__len_chromosome)
        end = 2 ** i - 1
        start = self.__mask - end
        return (partner.get_chromosome() & start) + (self.__chromosome & end)

    def __repr__(self):
        return format(self.__chromosome, '0' + str(self.__len_chromosome) +
                      'b') + '\t' + str(self.__chromosome)


class GeneticAlgorithm(object):
    '''Base class to run a genetic algorithm.'''
    __metaclass__ = abc.ABCMeta

    def __init__(self, pop_size, args, retain=0.2, random_select=0.05,
                 mutate=0.01, verbose=False):
        self.__pop_size = pop_size
        self.__args = args
        self.__retain = retain
        self.__random_select = random_select
        self.__mutate = mutate
        self._verbose = verbose
        self.__population = []

        while len(list(numpy.unique(numpy.array(self.__population)))) < \
                pop_size:
            self.__population.append(self.__get_individual())

    def run(self, max_iter=1024, max_tries=1024):
        '''Runs the genetic algorithm.'''
        for _ in range(max_iter):
            result = self.__evolve(max_tries)

            if result is not None:
                return result

        raise ValueError('Unable to optimise in ' + str(max_iter) +
                         ' iterations.')

    @abc.abstractmethod
    def _fitness(self, individual):
        '''Determine the fitness of an individual.'''
        return

    def __get_individual(self):
        '''Create a member of the population.'''
        return {k: _get_arg(args) for k, args in self.__args.iteritems()}

    def __evolve(self, max_tries):
        '''Performs one round of evolution.'''
        graded = sorted([(self._fitness(x), x) for x in self.__population])

        if graded[0][0] == 0:
            return graded[0][1]

        if self._verbose:
            print graded[0]

        graded = [x[1] for x in graded]
        retain_length = int(self.__pop_size * self.__retain)

        # Retain best and randomly add other individuals to promote genetic
        # diversity:
        self.__population = graded[:retain_length] + \
            [ind for ind in graded[retain_length:]
             if self.__random_select > random.random()]

        # Mutate some individuals:
        for individual in self.__population:
            if self.__mutate > random.random():
                key = random.choice(individual.keys())

                if key in self.__args:
                    individual[key] = _get_arg(self.__args[key])

        # Ensure uniqueness in population:
        self.__population = list(numpy.unique(numpy.array(self.__population)))

        self.__breed(max_tries)

    def __breed(self, max_tries):
        '''Breeds parents to create children.'''
        new_population = []
        tries = 0

        while len(new_population) < self.__pop_size:
            male = random.choice(self.__population)
            female = random.choice(self.__population)

            if male != female:
                pos = random.randint(0, len(male))

                male_parts = {k: male[k]
                              for i, k in enumerate(male.keys())
                              if i < pos}
                female_parts = {k: female[k]
                                for i, k in enumerate(female.keys())
                                if i >= pos}

                child = dict(male_parts.items() +
                             female_parts.items())

                new_population.append(child)
                new_population = \
                    list(numpy.unique(numpy.array(new_population)))

            tries += 1

            if tries == max_tries:
                raise ValueError('Unable to generate unique population.')

        self.__population = new_population


def _get_arg(args):
    '''Gets a random argument.'''
    return random.randint(args[0], args[1]) if isinstance(args, tuple) \
        else random.choice(args)



'''
synbiochem (c) University of Manchester 2015

synbiochem is licensed under the MIT License.

To view a copy of this license, visit <http://opensource.org/licenses/MIT/>.

@author:  neilswainston
'''
# pylint: disable=too-many-public-methods
import unittest

from synbiochem.utils import seq_utils


class CodonOptimiserTest(unittest.TestCase):
    '''Test class for CodonOptimiser.'''

    def test_get_codon_prob(self):
        '''Tests get_codon_prob method.'''
        cod_opt = seq_utils.CodonOptimiser('83333')
        self.assertAlmostEquals(0.46, cod_opt.get_codon_prob('CTG'), 2)

    def test_get_random_codon(self):
        '''Tests get_random_codon method.'''
        cod_opt = seq_utils.CodonOptimiser('83333')
        self.assertEquals('CTA', cod_opt.get_random_codon('L', ['CTG', 'TTA',
                                                                'CTT', 'TTG',
                                                                'CTC']))

    def test_get_random_codon_fail(self):
        '''Tests get_random_codon method.'''
        cod_opt = seq_utils.CodonOptimiser('83333')
        self.assertRaises(
            ValueError, cod_opt.get_random_codon, 'M', ['ATG'])


class Test(unittest.TestCase):
    '''Test class for sequence_utils.'''

    def test_get_codon_usage_organisms(self):
        '''Tests get_codon_usage_organisms() method.'''
        organisms = seq_utils.get_codon_usage_organisms()
        expected = {'\'Arthromyces ramosus\'': '5451',
                    'thiosulfate-reducing anaerobe SRL 4198': '267367'}
        self.assertDictContainsSubset(expected, organisms)

    def test_get_uniprot_values(self):
        '''Tests get_uniprot_values method.'''
        result = seq_utils.get_uniprot_values(['P19367', 'P42212'],
                                              ['organism-id',
                                               'protein names'], 1)

        expected = {'P19367': {'Organism ID': '9606',
                               'Protein names': ['Hexokinase-1',
                                                 'EC 2.7.1.1',
                                                 'Brain form hexokinase',
                                                 'Hexokinase type I',
                                                 'HK I']},
                    'P42212': {'Organism ID': '6100',
                               'Protein names': ['Green fluorescent protein']}}

        self.assertEquals(result, expected)

    def test_get_seq_by_melt_temp(self):
        '''Tests translate method.'''
        seq, _ = seq_utils.get_seq_by_melt_temp('agcgtgcgaagcgtgcgatcctcc', 70)
        self.assertEqual(seq, 'agcgtgcgaagcgtgcgatc')

    def test_translate(self):
        '''Tests translate method.'''
        results = seq_utils.translate('agcgtgcgatcc', min_prot_len=1)
        self.assertIn('ACD', [tokens[5] for tokens in results])

    def test_do_blast(self):
        '''Tests do_blast method.'''
        id_seq = {'test': seq_utils.get_random_dna(1024)}
        results = seq_utils.do_blast(id_seq, id_seq, evalue=10, word_size=4)

        alignments = []

        for result in results:
            for alignment in result.alignments:
                for hsp in alignment.hsps:
                    alignments.append(hsp)
                    print hsp

        self.assertGreater(len(alignments), 1)

if __name__ == "__main__":
    # import sys;sys.argv = ['', 'Test.testName']
    unittest.main()



'''
synbiochem (c) University of Manchester 2015

synbiochem is licensed under the MIT License.

To view a copy of this license, visit <http://opensource.org/licenses/MIT/>.

@author:  neilswainston
'''
# pylint: disable=too-many-public-methods
import unittest

from synbiochem.utils import struct_utils


class Test(unittest.TestCase):
    '''Test class for structure_utils.'''

    def test_get_seq_struct(self):
        '''Tests isclose method.'''
        seq_structs = struct_utils.get_seq_structs(['1DWI'])
        self.assertTrue(all([len(v[0]) == len(v[1])
                             for v in seq_structs.values()]))


if __name__ == "__main__":
    # import sys;sys.argv = ['', 'Test.testName']
    unittest.main()



'''
synbiochem (c) University of Manchester 2015

synbiochem is licensed under the MIT License.

To view a copy of this license, visit <http://opensource.org/licenses/MIT/>.

@author:  neilswainston
'''
# pylint: disable=no-member
# pylint: disable=too-few-public-methods
# pylint: disable=too-many-public-methods
import random
import unittest

import synbiochem.optimisation.gen_alg as gen_alg


class TestChromosome(unittest.TestCase):
    '''Class to test Chromosome class.'''

    def test_breed(self):
        '''Test for breed method.'''
        length = 16
        chrom1 = gen_alg.Chromosome(0, length)
        chrom2 = gen_alg.Chromosome(2 ** length - 1, length)

        for _ in range(1024):
            chrom1.breed(chrom2)

        self.assertEqual(
            chrom1.get_chromosome() + chrom2.get_chromosome(), 2 ** length - 1)


class TestSumGeneticAlgorithm(unittest.TestCase):
    '''Class to test SumGeneticAlgorithm class.'''

    def test_run(self):
        '''Test for run method.'''
        args = dict(enumerate([[5, 10]] + [(random.randint(0, 20),
                                            random.randint(80, 100))
                                           for _ in range(10)]))
        target = 321
        genetic_algorithm = SumGeneticAlgorithm(100, args, target)
        self.assertEqual(sum(genetic_algorithm.run(100000).values()), target)

    def test_run_error(self):
        '''Test for run method.'''
        args = dict(enumerate([[5, 10]] + [(random.randint(0, 20),
                                            random.randint(80, 100))
                                           for _ in range(10)]))
        target = 936073
        genetic_algorithm = SumGeneticAlgorithm(100, args, target)
        self.assertRaises(ValueError, genetic_algorithm.run)


class SumGeneticAlgorithm(gen_alg.GeneticAlgorithm):
    '''Class to run a genetic algorithm.
    Basic implementation involves calculating a set of numbers that sum to
    a given target.'''

    def __init__(self, pop_size, args, target):
        super(SumGeneticAlgorithm, self).__init__(pop_size, args)
        self.__target = target

    def _fitness(self, individual):
        '''Determine the fitness of an individual.'''
        return abs(self.__target - sum(individual.values()))



'''
synbiochem (c) University of Manchester 2015

synbiochem is licensed under the MIT License.

To view a copy of this license, visit <http://opensource.org/licenses/MIT/>.

@author:  neilswainston
'''
# pylint: disable=no-member
import gurobipy


def isclose(value_1, value_2, rel_tol=1e-09, abs_tol=0.0):
    '''Compares floating point numbers.'''
    return abs(value_1 - value_2) <= max(rel_tol * max(abs(value_1),
                                                       abs(value_2)), abs_tol)


def linprog(c_vector, a_eq, b_eq, bounds):
    '''Solve linear programming problem.'''
    model = gurobipy.Model()
    model.Params.OutputFlag = 0  # Silent mode

    # Create variables:
    for obj, bound in zip(c_vector, bounds):
        model.addVar(bound[0], bound[1], obj, gurobipy.GRB.CONTINUOUS)

    # Integrate new variables:
    model.update()

    # Add constraints:
    for aaa, bbb in zip(a_eq, b_eq):
        model.addConstr(gurobipy.LinExpr(aaa, model.getVars()),
                        gurobipy.GRB.EQUAL,
                        bbb)

    model.optimize()

    return model.status == gurobipy.GRB.status.OPTIMAL, \
        [var.X for var in model.getVars()] \
        if model.status == gurobipy.GRB.status.OPTIMAL else None



'''
synbiochem (c) University of Manchester 2015

synbiochem is licensed under the MIT License.

To view a copy of this license, visit <http://opensource.org/licenses/MIT/>.

@author:  neilswainston
'''
# pylint: disable=too-many-public-methods
import unittest

import synbiochem.utils.chem_utils as chm_util


class Test(unittest.TestCase):
    '''Test class for chemical_utils.'''

    def test_get_molecular_mass(self):
        '''Tests get_molecular_mass method.'''
        self.assertAlmostEqual(chm_util.get_molecular_mass('H2O'),
                               18.010564684)

    def test_get_elem_comp(self):
        '''Tests get_elem_comp method.'''
        self.assertEqual(chm_util.get_elem_comp('Fe4S'), {'Fe': 4, 'S': 1})

    def test_parse_equation(self):
        '''Tests parse_equation method.'''
        eqn = '5.6 Fe4S + -3.2 water = 17.8 SiO2'
        self.assertEqual(chm_util.parse_equation(eqn),
                         [['Fe4S', -5.6], ['water', 3.2], ['SiO2', 17.8]])

    def test_parse_equation_error(self):
        '''Tests parse_equation method (with error).'''
        eqn = '5.6 Fe4S + -3.2 water = n+m SiO2'
        self.assertRaises(ValueError, chm_util.parse_equation, eqn)

    def test_balance_unbalanced(self):
        '''Tests get_elem_comp method for unbalanced reaction.'''
        unbalanced = [('CO2', 0, -1.0, 'CO2'),
                      ('C5H7O4', -1, -1.0, 'C5H7O4'),
                      ('C3H3O3', -1, 1.0, 'C3H3O3')]
        is_balanced, was_balanced, balanced_def = chm_util.balance(unbalanced)

        self.assertTrue(is_balanced)
        self.assertFalse(was_balanced)
        self.assertEqual(sorted([('C3H3O3', -1, 2.0, 'C3H3O3'),
                                 ('C5H7O4', -1, -1.0, 'C5H7O4'),
                                 ('CO2', 0, -1.0, 'CO2'),
                                 ('H', 1, 1.0, 'CHEBI:24636')]),
                         sorted(balanced_def))

    def test_balance_balanced(self):
        '''Tests get_elem_comp method for balanced reaction.'''
        balanced = [('C5H7O4', -1, -1.0, 'C5H7O4'),
                    ('H', 1, 1.0, 'H'),
                    ('C3H3O3', -1, 2.0, 'C3H3O3'),
                    ('CO2', 0, -1.0, 'CO2')]
        is_balanced, was_balanced, balanced_def = chm_util.balance(balanced)

        self.assertTrue(is_balanced)
        self.assertTrue(was_balanced)
        self.assertEqual(sorted(balanced), sorted(balanced_def))

    def test_balance_problem(self):
        '''Tests get_elem_comp method for problematic reaction.'''
        problem = [('C144H238N2O57P2', -2.0, -1.0, 'C144H238N2O57P2'),
                   ('C86H142O9P', -1.0, -1.0, 'C86H142O9P'),
                   ('H', 1.0, 1.0, 'H'),
                   ('C150H248N2O62P2', -2.0, 1.0, 'C150H248N2O62P2'),
                   ('C80H131O4P', -2.0, 1.0, 'C80H131O4P')]

        is_balanced, was_balanced, balanced_def = chm_util.balance(
            problem, optional_comp=[])

        self.assertTrue(is_balanced)
        self.assertTrue(was_balanced)
        self.assertEqual(sorted(problem), sorted(balanced_def))

if __name__ == "__main__":
    # import sys;sys.argv = ['', 'Test.testName']
    unittest.main()



'''
synbiochem (c) University of Manchester 2015

synbiochem is licensed under the MIT License.

To view a copy of this license, visit <http://opensource.org/licenses/MIT/>.

@author:  neilswainston
'''
# pylint: disable=too-many-arguments
from itertools import izip_longest, product
import collections
import os
import random
import re
import sys

from Bio import SeqUtils
from Bio.PDB.PDBParser import PDBParser
from Bio.PDB.Polypeptide import PPBuilder
from matplotlib.colors import LinearSegmentedColormap
import pylab

import numpy as np
import scipy.spatial.distance as dist
import synbiochem.utils.io_utils as io_utils


_DIR = 'structure_utils'
_PDB_DIR = 'pdb'
_SEC_STRUCT = 'EBHIGTSL '


def get_pdb_ids(max_ids=None, local_only=False):
    '''Returns PDB ids.'''
    if local_only:
        # Returns local PDB ids.
        pdb_dir = os.path.join(os.path.expanduser('~'), _DIR, _PDB_DIR)
        ids = [filename[:-4].upper()
               for _, _, files in os.walk(pdb_dir)
               for filename in files if filename.endswith('.pdb')]
    else:
        # Returns all PDB ids.
        source_url = 'http://www.uniprot.org/uniprot/?query=database:pdb' \
            + '&format=tab&columns=id,database(PDB)'
        target_filename = os.path.join(os.path.expanduser('~'), _DIR,
                                       'pdb_ids.txt')

        with open(io_utils.get_file(source_url, target_filename)) as fle:
            ids = [x for line in fle
                   for x in line.split()[1].split(';')
                   if len(x) > 0 and x != 'Cross-reference']

    return ids if max_ids is None \
        else random.sample(ids, min(len(ids), max_ids))


def get_seq_structs(pdb_ids=None):
    '''Returns sequence and structure.'''
    seq_structs = {}
    pdb_ids = sorted(pdb_ids) if pdb_ids is not None else None
    in_field = False
    tokens = ()
    str_data = ''

    source_url = 'http://www.rcsb.org/pdb/files/ss.txt'
    target_filename = os.path.join(os.path.expanduser('~'), _DIR,
                                   'ss.txt')

    with open(io_utils.get_file(source_url, target_filename)) as fle:
        for line in fle:
            if line.startswith('>'):
                pdb_id = re.search('(?<=\\>)[^:]*', line).group(0)

                if pdb_ids is None or pdb_id in pdb_ids:
                    if in_field:
                        if tokens[:2] not in seq_structs:
                            seq_structs[tokens[:2]] = [None, None]

                        seq_structs[tokens[:2]][0 if tokens[2] == 'sequence'
                                                else 1] = str_data
                        str_data = ''

                    tokens = tuple(re.split('>|:', line.strip())[1:])
                    in_field = True

                elif in_field:
                    if tokens[:2] not in seq_structs:
                        seq_structs[tokens[:2]] = [None, None]

                    seq_structs[tokens[:2]][0 if tokens[2] == 'sequence'
                                            else 1] = str_data
                    str_data = ''

                    in_field = False
                    tokens = ()
                    str_data = ''

            elif in_field:
                str_data += line[:-1]

    return {key: value for key, value in seq_structs.iteritems()
            if all(val is not None for val in value)}


def get_pep_structs(struct_set, length):
    '''Returns sequence and structure.'''
    for struct in struct_set:
        assert struct in _SEC_STRUCT

    return [pep for struct in struct_set for pep in _get_peps(struct, length)]


def get_sequences(pdb_id, chain=None):
    '''Gets the sequences in a PDB file.'''
    return [SeqUtils.seq1(''.join([residue.get_resname()
                                   for residue in chn
                                   if 'CA' in residue.child_dict]))
            for chn in get_structure(pdb_id).get_chains()
            if chain is None or chain == chn.get_id()]


def get_structure(pdb_id):
    '''Returns a PDB structure.'''
    source_url = 'http://www.rcsb.org/pdb/files/' + pdb_id + '.pdb'
    target_filename = os.path.join(os.path.expanduser('~'), _DIR, _PDB_DIR,
                                   pdb_id + '.pdb')

    with open(io_utils.get_file(source_url, target_filename)) as pdb_file:
        parser = PDBParser(QUIET=True)
        return parser.get_structure(pdb_id, pdb_file.name)


def calc_proximities(pdb_id):
    '''Calculates residue proximities from PDB file.'''
    all_proximities = []
    structure = get_structure(pdb_id)
    chains = [c for c in structure.get_chains()]

    for chn in chains:
        res_coords = []

        for residue in chn:
            if 'CA' in residue.child_dict:
                res_coords.append([atom.get_coord()
                                   for atom_id, atom
                                   in residue.child_dict.iteritems()
                                   if atom_id
                                   not in ['C', 'N', 'O', 'OXT']])

        residues = [residue.get_id()[1]
                    for residue in chn
                    if 'CA' in residue.child_dict]

        chn_prox = collections.defaultdict(list)

        for res1, res2 in product(zip(residues, res_coords), repeat=2):
            chn_prox[res1[0]].append(np.amin(dist.cdist(res1[1], res2[1],
                                                        'euclidean')))

        all_proximities.append([chn_prox.keys(), chn_prox.values()])

    return all_proximities


def get_phi_psi_data(pdb_id, chain=None):
    '''Gets phi and phi angle data.'''
    builder = PPBuilder()
    return [polypep.get_phi_psi_list()
            for model in get_structure(pdb_id)
            for chn in model
            if chain is None or chain == chn.get_id()
            for polypep in builder.build_peptides(chn)]


def plot_proximities(pdb_id, groupings=None, group_len=0):
    '''Plots proximity plot(s).'''
    all_proximities = _group_proximities(calc_proximities(pdb_id), groupings,
                                         group_len)

    plot_format = 'png'

    for idx, proximities in enumerate(all_proximities):
        name = pdb_id + '_' + str(idx + 1)
        _plot(proximities[0], proximities[1], name + '.' + plot_format,
              plot_format, name + ' proximity plot')


def _plot(residues, values, plot_filename, plot_format, title, max_value=None):
    '''Plots 3d matrix values.'''
    fig = pylab.figure()
    sub_plot = fig.add_subplot(111)

    cmap = LinearSegmentedColormap.from_list(name='name', colors=['g', 'w'],
                                             N=10)

    cax = sub_plot.imshow(values, interpolation='nearest', cmap=cmap,
                          extent=[residues[0], residues[-1],
                                  residues[-1], residues[0]])
    cax.set_clim(0.0, max_value)
    sub_plot.set_title(title)

    # Add colorbar, make sure to specify tick locations to match desired tick
    # labels
    min_val = np.min(values)
    max_val = np.max(values)
    cbar = fig.colorbar(cax, ticks=[min_val, max_val])
    cbar.set_ticks([min_val, max_val])

    pylab.savefig(plot_filename, format=plot_format)


def _get_peps(struct, length):
    '''Gets n-mers of length, matching a given structure.'''
    directory = os.path.join(os.path.expanduser('~'), _DIR, str(length))
    filename = _get_struct_filename(directory, struct)

    if not os.path.isfile(filename):
        _write_peps(length, directory)

    with open(filename) as fle:
        return [line.split()[0] for line in fle]


def _write_peps(length, directory):
    '''Writes n-mers of length, matching a given structure, to file.'''
    assert length % 2 == 1

    pep_structs = {}

    for seq_struct in get_seq_structs().values():
        seq = seq_struct[0]
        struct = seq_struct[1]

        for i in range(len(seq) - length + 1):
            pep_seq = seq[i:i + length]
            pep_struct = struct[i:i + length]

            if pep_seq not in pep_structs:
                pep_structs[pep_seq] = set([pep_struct])
            else:
                pep_structs[pep_seq].add(pep_struct)

    classified_peps = _get_classified_peps(pep_structs, length / 2)
    _write_classified_peps(classified_peps, directory)


def _get_classified_peps(pep_structs, middle_index):
    '''Gets peptides classified by middle residue secondary structure,
    e.g. H, S, B, etc.'''
    classified_peps = {key: [] for key in list(_SEC_STRUCT) + ['unknown']}

    for pep, structs in pep_structs.iteritems():
        structs = list(structs)
        clss = 'unknown' if len(structs) > 1 else structs[0][middle_index]
        classified_peps[clss].append([pep, structs])

    return classified_peps


def _write_classified_peps(classified_peps, directory):
    '''Writes classified peptides to file.'''
    if not os.path.exists(directory):
        os.makedirs(directory)

    for clss, peps in classified_peps.iteritems():
        with open(_get_struct_filename(directory, clss), 'w+') as outfile:
            for pep in peps:
                outfile.write('\t'.join([str(val) for val in pep]) + '\n')


def _get_struct_filename(directory, struct):
    '''Gets filename for a given secondary structure class.'''
    # struct = struct if struct is not ' ' else 'blank'
    return os.path.join(directory, struct + '.xls')


def _group_proximities(proximities, groupings, group_len=0):
    '''Groups proximities.'''
    if groupings is None and group_len < 1:
        return proximities

    group_proxs = []

    for chain_prox in proximities:
        len_prox = (max(chain_prox[0]) + 1)

        if group_len > 0:
            groupings = [[item for item in vals if item is not None]
                         for vals in izip_longest(*[iter(range(len_prox))] *
                                                  group_len)]

        prox_list = [[float('NaN')] * len_prox] * len_prox

        for prox, value in zip(*chain_prox):
            prox_list[prox] = [float('NaN')] * len_prox

            for idx, res in enumerate(chain_prox[0]):
                prox_list[prox][res] = value[idx]

        group_proxs.append([range(len(groupings)), _get_group_prox(groupings,
                                                                   prox_list)])

    return group_proxs


def _get_group_prox(groupings, prox_list):
    group_prox = [None] * len(groupings)

    for idx1, group1 in enumerate(groupings):
        group_prox[idx1] = [None] * len(groupings)

        for idx2, group2 in enumerate(groupings):
            group_prox[idx1][idx2] = np.mean([prox_list[pair[0]][pair[1]]
                                              for pair in product(group1,
                                                                  group2)])

    return group_prox


def main(args):
    '''main method'''
    print
    plot_proximities(args[0])

if __name__ == '__main__':
    main(sys.argv[1:])



'''
synbiochem (c) University of Manchester 2015

synbiochem is licensed under the MIT License.

To view a copy of this license, visit <http://opensource.org/licenses/MIT/>.

@author:  neilswainston
'''
# pylint: disable=no-member
# pylint: disable=too-many-arguments
# pylint: disable=too-many-locals
from subprocess import call, Popen
import collections
import itertools
import operator
import os
import random
import tempfile
import urllib
import urllib2

from Bio.Blast import NCBIXML
from Bio.Data import CodonTable, IUPACData
from Bio.Seq import Seq
from Bio.SeqUtils.MeltingTemp import Tm_NN
from requests.exceptions import ConnectionError
from synbiochem.biochem4j import taxonomy
import numpy

import regex as re


NUCLEOTIDES = ['A', 'C', 'G', 'T']

AA_CODES = {'Ala': 'A',
            'Cys': 'C',
            'Asp': 'D',
            'Glu': 'E',
            'Phe': 'F',
            'Gly': 'G',
            'His': 'H',
            'Ile': 'I',
            'Lys': 'K',
            'Leu': 'L',
            'Met': 'M',
            'Asn': 'N',
            'Pro': 'P',
            'Gln': 'Q',
            'Arg': 'R',
            'Ser': 'S',
            'Thr': 'T',
            'Val': 'V',
            'Trp': 'W',
            'Tyr': 'Y',
            'End': '*'}

CODONS = {'A': [['G', 'C', 'ACGT']],
          'C': [['T', 'G', 'CT']],
          'D': [['G', 'A', 'CT']],
          'E': [['G', 'A', 'AG']],
          'F': [['T', 'T', 'CT']],
          'G': [['G', 'G', 'ACGT']],
          'H': [['C', 'A', 'CT']],
          'I': [['A', 'T', 'ACT']],
          'K': [['A', 'A', 'AG']],
          'L': [['C', 'T', 'ACGT'], ['T', 'T', 'AG']],
          'M': [['A', 'T', 'G']],
          'N': [['A', 'A', 'CT']],
          'P': [['C', 'C', 'ACGT']],
          'Q': [['C', 'A', 'AG']],
          'R': [['C', 'G', 'ACGT'], ['A', 'G', 'AG']],
          'S': [['T', 'C', 'ACGT'], ['A', 'G', 'CT']],
          'T': [['A', 'C', 'ACGT']],
          'V': [['G', 'T', 'ACGT']],
          'W': [['T', 'G', 'G']],
          'Y': [['T', 'A', 'CT']],
          'Stop': [['T', 'A', 'AG'], ['T', 'G', 'A']]}

NUCL_CODES = {
    'A': 'A',
    'C': 'C',
    'G': 'G',
    'T': 'T',
    'AG': 'R',
    'CT': 'Y',
    'CG': 'S',
    'AT': 'W',
    'GT': 'K',
    'AC': 'M',
    'CGT': 'B',
    'AGT': 'D',
    'ACT': 'H',
    'ACG': 'V',
    'ACGT': 'N',
}

INV_NUCL_CODES = {val: key for key, val in NUCL_CODES.items()}

# KD Hydrophobicity, EIIP, Helix, Sheet, Turn
AA_PROPS = {
    'A': [1.8, -0.0667, 32.9, -23.6, -41.6],
    'R': [-4.5, 0.2674, 0, -6.2, -5.1],
    'N': [-3.5, -0.2589, -24.8, -41.6, 44.5],
    'D': [-3.5, 0.4408, 5.8, -41.6, 37.8],
    'C': [2.5, 0.1933, -5.1, 6.8, 17.4],
    'Q': [-3.5, 0.1545, 11.3, 0, -2],
    'E': [-3.5, -0.2463, 36.5, -67.3, -30.1],
    'G': [-0.4, -0.2509, -46.2, -13.9, 44.5],
    'H': [-3.2, -0.1414, 11.3, -18.6, -5.1],
    'I': [4.5, -0.2794, -1, 45.1, -75.5],
    'L': [3.8, -0.2794, 26.2, 15.7, -52.8],
    'K': [-3.9, -0.0679, 19.1, -31.5, 1],
    'M': [1.9, 0.1899, 27.8, 1, -51.1],
    'F': [2.8, 0.26, 10.4, 20.7, -51.1],
    'P': [-1.6, -0.1665, -59.8, -47.8, 41.9],
    'S': [-0.8, 0.1933, -32.9, -6.2, 35.8],
    'T': [-0.7, 0.2572, -24.8, 28.5, -4.1],
    'W': [-0.9, 0.0331, 3, 21.5, -4.1],
    'Y': [-1.3, 0.0148, -31.5, 27, 13.1],
    'V': [4.2, -0.2469, -3, 49.5, -69.3]
}

NUM_AA_PROPS = len(AA_PROPS['A'])

NA = 'NA'
K = 'K'
TRIS = 'TRIS'
MG = 'MG'
DNTP = 'DNTP'

START_CODON_PATT = '[ACGT]TG'


def get_aa_props(all_sequences, scale=(0.1, 0.9)):
    '''Returns input data for machine-learning problems.'''
    scaled = __scale(scale)
    mean_value = numpy.mean([x for sublist in scaled.values()
                             for x in sublist])
    return [list(itertools.chain.from_iterable([scaled[am_acid]
                                                if am_acid in scaled
                                                else [mean_value] *
                                                len(scaled['A'])
                                                for am_acid in sequences]))
            for sequences in all_sequences]


def __scale(scale):
    '''Scale amino acid properties.'''
    scaled = collections.defaultdict(list)

    for i in range(NUM_AA_PROPS):
        props = {key: value[i] for key, value in AA_PROPS.iteritems()}
        min_val, max_val = min(props.values()), max(props.values())

        for key, value in props.iteritems():
            scaled[key].append(scale[0] + (scale[1] - scale[0]) *
                               (value - min_val) / (max_val - min_val))

    return scaled

__DEFAULT_REAG_CONC = {NA: 0.05, K: 0, TRIS: 0, MG: 0.01, DNTP: 0}


def get_codon_usage_organisms(expand=False):
    '''Gets name to taxonomy id dictionary of available codon usage tables.'''
    destination = os.path.join(os.path.expanduser('~'), 'codon')
    filename = 'expand.txt' if expand else 'normal.txt'
    filepath = os.path.join(destination, filename)

    if not os.path.exists(filepath):
        # Download:
        if not os.path.exists(destination):
            os.makedirs(destination)

        url = 'ftp://ftp.kazusa.or.jp/pub/codon/current/species.table'
        tmp = tempfile.NamedTemporaryFile(delete=False)
        urllib.urlretrieve(url, tmp.name)

        # Read:
        codon_orgs = _read_codon_usage_orgs_file(tmp.name)

        # Expand:
        if expand:
            _expand_codon_usage_orgs(codon_orgs)

        # Save:
        _write_codon_usage_orgs_file(codon_orgs, filepath)

        return codon_orgs
    else:
        return _read_codon_usage_orgs_file(filepath)


class CodonOptimiser(object):
    '''Class to support codon optimisation.'''

    def __init__(self, taxonomy_id):
        self.__taxonomy_id = taxonomy_id
        self.__aa_to_codon_prob = self.__get_codon_usage()
        self.__codon_prob = {item[0]: item[1]
                             for lst in self.__aa_to_codon_prob.values()
                             for item in lst}

        self.__codon_to_w = {}

        for key in self.__aa_to_codon_prob:
            aa_dict = dict([(a, b / self.__aa_to_codon_prob[key][0][1])
                            for a, b in self.__aa_to_codon_prob[key]])
            self.__codon_to_w.update(aa_dict)

    def get_codon_prob(self, codon):
        '''Gets the codon probability.'''
        return self.__codon_prob[codon]

    def optimise(self, protein_seqs, max_repeat_nuc=float('inf')):
        '''Codon optimises the supplied protein sequences.'''
        optimised_seqs = []
        max_attempts = 1000
        attempts = 0

        for protein_seq in protein_seqs:
            while True:
                attempts += 1

                if attempts > max_attempts:
                    raise ValueError('Unable to optimise sequence. ' +
                                     'Greater than ' + str(max_repeat_nuc) +
                                     ' repeating nucleotides.')

                optimised_seq = self.get_codon_optim_seq(protein_seq)

                if is_valid(optimised_seq, max_repeat_nuc):
                    optimised_seqs.append(optimised_seq)
                    break

        return optimised_seqs

    def get_codon_optim_seq(self, protein_seq, excl_codons=None,
                            invalid_pattern=None, max_attempts=100,
                            tolerant=False):
        '''Returns a codon optimised DNA sequence.'''
        if invalid_pattern is None:
            return ''.join([self.get_random_codon(aa)
                            for aa in protein_seq])
        else:
            attempts = 0
            seq = ''
            i = 0
            blockage_i = -1
            inv_patterns = 0

            while attempts < max_attempts:
                amino_acid = protein_seq[i]
                new_seq = seq + self.get_random_codon(amino_acid, excl_codons)

                invalids = [invalid.start()
                            for invalid in re.finditer(invalid_pattern,
                                                       new_seq,
                                                       overlapped=True)]
                if len(invalids) == inv_patterns or \
                        (attempts == max_attempts - 1 and tolerant):

                    if i == blockage_i:
                        if attempts == max_attempts - 1:
                            inv_patterns = inv_patterns + 1

                        attempts = 0

                    seq = new_seq

                    if i == len(protein_seq) - 1:
                        return seq

                    i += 1
                else:
                    blockage_i = max(i, blockage_i)
                    i = max(0, (invalids[-1] / 3) - 1)
                    seq = seq[:i * 3]
                    attempts += 1

            raise ValueError('Unable to generate codon-optimised sequence.')

    def get_cai(self, dna_seq):
        '''Gets the CAI for a given DNA sequence.'''
        cai = 0

        for i in range(0, len(dna_seq), 3):
            cai += self.__codon_to_w[dna_seq[i:i + 3]]

        return cai / (len(dna_seq) / 3)

    def mutate(self, protein_seq, dna_seq, mutation_rate):
        '''Mutate a protein-encoding DNA sequence according to a
        supplied mutation rate.'''
        return ''.join([self.get_random_codon(amino_acid)
                        if random.random() < mutation_rate
                        else dna_seq[3 * i:3 * (i + 1)]
                        for i, amino_acid in enumerate(protein_seq)])

    def get_all_rev_trans(self, aa_seq):
        '''Returns all reverse translations of amino acid sequence.'''
        codons = [self.get_all_codons(aa) for aa in aa_seq]
        return [''.join(t) for t in list(itertools.product(*codons))]

    def get_all_codons(self, amino_acid):
        '''Returns all codons for a given amino acid.'''
        return [t[0] for t in self.__aa_to_codon_prob[amino_acid]]

    def get_random_codon(self, amino_acid, excl_codons=None):
        '''Returns a random codon for a given amino acid,
        based on codon probability from the codon usage table.'''
        if excl_codons is None:
            excl_codons = []

        codon_usage = [codon_usage
                       for codon_usage in self.__aa_to_codon_prob[amino_acid]
                       if codon_usage[0] not in excl_codons]

        if len(codon_usage) == 0:
            raise ValueError('No codons available for ' + amino_acid +
                             ' after excluding ' + str(excl_codons))

        while True:
            rand = random.random()
            cumulative_prob = 0

            for codon, prob in iter(reversed(codon_usage)):
                cumulative_prob += prob

                if cumulative_prob > rand:
                    return codon

    def __get_codon_usage(self):
        '''Gets the codon usage table for a given taxonomy id.'''
        aa_to_codon_prob = {aa_code: {} for aa_code in AA_CODES.values()}

        url = 'http://www.kazusa.or.jp/codon/cgi-bin/showcodon.cgi?species=' \
            + self.__taxonomy_id + '&aa=1&style=GCG'

        in_codons = False

        for line in urllib2.urlopen(url):
            if line == '<PRE>\n':
                in_codons = True
            elif line == '</PRE>\n':
                break
            elif in_codons:
                values = re.split('\\s+', line)

                if values[0] in AA_CODES:
                    codon_prob = aa_to_codon_prob[AA_CODES[values[0]]]
                    codon_prob[values[1]] = float(values[3])

        aa_to_codon_prob.update((x, _scale(y))
                                for x, y in aa_to_codon_prob.items())

        return aa_to_codon_prob


def get_minimum_free_energy(sequences):
    '''Returns minimum free energy of supplied DNA / RNA sequences.'''
    with open(tempfile.NamedTemporaryFile(), 'w') as input_file, \
            open(tempfile.NamedTemporaryFile(), 'w') as output_file:

        for i, sequence in enumerate(sequences):
            input_file.write('>Seq' + str(i) + '\n' + sequence + '\n')
            input_file.flush()

        input_file.close()

        proc = Popen('RNAfold',
                     stdin=open(input_file.name),
                     stdout=output_file)

        proc.wait()

        _cleanup(os.getcwd(), 'Seq\\d+_ss.ps')

        mfes = []
        pattern = re.compile(r'[+-]?\d+\.\d+')

        with open(output_file.name) as out_file:
            for line in out_file.readlines():
                src = pattern.search(line)

                if src:
                    mfes.append(float(src.group()))

        return mfes


def get_random_dna(length, max_repeat_nuc=float('inf'), invalid_patterns=None):
    '''Returns a random sequence of DNA of the supplied length,
    while adhering to a maximum number of repeating nucleotides.'''
    max_attempts = 1000
    attempts = 0

    if invalid_patterns is None:
        invalid_patterns = []

    while True:
        attempts += 1

        if attempts > max_attempts:
            raise ValueError('Unable to optimise sequence. ' +
                             'Greater than ' + str(max_repeat_nuc) +
                             ' repeating nucleotides.')

        random_dna = _get_random_dna(length)

        valid = True
        for invalid_pattern in invalid_patterns:

            if len(re.findall(invalid_pattern, random_dna, overlapped=True)) \
                    > 0:
                valid = False

        if valid and is_valid(random_dna, max_repeat_nuc):
            return random_dna

    return None


def get_random_aa(length, insertions=False):
    '''Returns a random amino acid sequence of the supplied length.'''
    dictionary = list(AA_PROPS.keys()) + (['.'] if insertions else [])
    return ''.join(random.choice(dictionary) for _ in range(length))


def mutate_seq(seq, mutations=1, alphabet=None):
    '''Mutates sequence.'''
    if alphabet is None:
        alphabet = NUCLEOTIDES

    seq_new = seq

    for _ in range(mutations):
        move = random.random()
        pos = int(random.random() * len(seq))
        base = random.choice(alphabet)

        # Insert:
        if move < 0.1:
            seq_new = seq_new[1:pos + 1] + base + seq_new[pos + 1:]

        # Delete:
        elif move < 0.2:
            seq_new = base + seq_new[:pos] + seq_new[pos + 1:]

        # Replace:
        else:
            seq_new = seq_new[:pos] + base + seq_new[pos + 1:]

    return seq_new


def get_melting_temp(dna1, dna2=None, reag_concs=None, strict=True):
    '''Calculates melting temperarure of DNA sequence against its
    complement, or against second DNA sequence using Nearest-Neighbour
    method.'''
    assert len(dna1) > 1

    reagent_concs = __DEFAULT_REAG_CONC

    if reag_concs is not None:
        reagent_concs.update(reag_concs)

    reagent_conc = {k: v * 1000 for k, v in reagent_concs.iteritems()}
    dnac1 = 30

    return Tm_NN(dna1, check=True, strict=strict, c_seq=dna2, shift=0,
                 Na=reagent_conc[NA], K=reagent_conc[K],
                 Tris=reagent_conc[TRIS], Mg=reagent_conc[MG],
                 dNTPs=reagent_conc[DNTP],
                 dnac1=dnac1, dnac2=dnac1, selfcomp=dna2 is None,
                 saltcorr=7)


def get_seq_by_melt_temp(seq, target_melt_temp, forward=True,
                         terminii=None,
                         reagent_concs=None,
                         tol=0.025):
    '''Returns a subsequence close to desired melting temperature.'''
    if terminii is None:
        terminii = ['A', 'C', 'G', 'T']
    else:
        terminii = [term.upper() for term in terminii]

    best_delta_tm = float('inf')
    best_subseq = ''
    best_melt_temp = float('NaN')
    in_tol = False

    for i in range(1, len(seq)):
        subseq = seq[:(i + 1)] if forward else seq[-(i + 1):]
        melt_temp = get_melting_temp(subseq, None, reagent_concs)

        if subseq[-1 if forward else 0].upper() in terminii:
            delta_tm = abs(melt_temp - target_melt_temp)

            if delta_tm / target_melt_temp < tol:
                in_tol = True

                if delta_tm < best_delta_tm:
                    best_delta_tm = delta_tm
                    best_subseq = subseq
                    best_melt_temp = melt_temp
            elif in_tol:
                return best_subseq, best_melt_temp

    raise ValueError('Unable to get sequence of required melting temperature')


def get_rand_seq_by_melt_temp(target_melt_temp, forward=True,
                              reagent_concs=None, init_len=1000,
                              max_repeat_nuc=float('inf')):
    '''Returns a random close to desired melting temperature.'''
    seq = get_random_dna(init_len, max_repeat_nuc)

    for i in range(1, len(seq)):
        subseq = seq[:(i + 1)] if forward else seq[-(i + 1):]
        melt_temp = get_melting_temp(subseq, None, reagent_concs)

        if melt_temp > target_melt_temp:
            return subseq, melt_temp

    raise ValueError('Unable to get sequence of required melting temperature')


def is_valid(dna_seq, max_repeat_nuc):
    '''Checks whether a DNA sequence is valid, in terms of a supplied maximum
    number of repeating nucleotides.'''
    nuc_count = 0
    prev_nuc = ''

    for nuc in dna_seq:
        if prev_nuc == nuc:
            nuc_count += 1
        else:
            prev_nuc = nuc
            nuc_count = 1

        if nuc_count > max_repeat_nuc:
            return False

    return True


def get_uniprot_values(uniprot_ids, fields, batch_size=64, verbose=False):
    '''Gets dictionary of ids to values from Uniprot.'''
    values = {}

    for i in xrange(0, len(uniprot_ids), batch_size):

        if verbose:
            print 'seq_utils: getting Uniprot values ' + str(i) + ' - ' + \
                str(min(i + batch_size, len(uniprot_ids))) + ' / ' + \
                str(len(uniprot_ids))

        batch = uniprot_ids[i:min(i + batch_size, len(uniprot_ids))]
        query = '+or+'.join(['id:' + uniprot_id for uniprot_id in batch])
        url = 'http://www.uniprot.org/uniprot/?query=' + query + \
            '&format=tab&columns=id,' + ','.join([urllib.quote(field)
                                                  for field in fields])

        _parse_uniprot_data(url, values)

    return values


def search_uniprot(name, fields, limit=128):
    '''Gets dictionary of ids to values from Uniprot.'''
    values = {}

    url = 'http://www.uniprot.org/uniprot/?query=name:' + name + \
        '&limit=' + str(limit) + \
        '&format=tab&columns=id,' + ','.join([urllib.quote(field)
                                              for field in fields])

    _parse_uniprot_data(url, values)

    return values


def count_pattern(strings, pattern, both_strands=True):
    '''Counts pattern in string of list of strings.'''
    forward = _count_pattern(strings, pattern)

    if both_strands:
        return forward + _count_pattern(get_rev_comp(strings), pattern)
    else:
        return forward


def get_hamming(str1, str2):
    '''Returns Hamming distance for two sequences, which are assumed to be of
    the same length.'''
    return sum(itertools.imap(operator.ne, str1, str2))


def get_rev_comp(seq):
    '''Returns reverse complement of sequence.'''
    seq = Seq(seq)
    return str(seq.reverse_complement())


def get_comp(seq):
    '''Returns complement of sequence.'''
    seq = Seq(seq)
    return str(seq.complement())


def do_blast(id_seqs_subjects, id_seqs_queries, program='blastn',
             dbtype='nucl', evalue=1.0, word_size=28):
    '''Performs BLAST of query sequences against subject sequences.'''

    db_filename = write_fasta(id_seqs_subjects)
    query_filename = write_fasta(id_seqs_queries)
    result_file = tempfile.NamedTemporaryFile(prefix='blast_result_',
                                              suffix='.xml',
                                              delete=False)
    call(['makeblastdb',
          '-in', db_filename,
          '-out', db_filename,
          '-dbtype', dbtype])

    call([program,
          '-query', query_filename,
          '-db', db_filename,
          '-out', result_file.name,
          '-evalue', str(evalue),
          '-word_size', str(word_size),
          '-outfmt', '5'])

    return NCBIXML.parse(open(result_file.name))


def write_fasta(id_seqs, filename=None):
    '''Writes a fasta file.'''
    if filename is None:
        temp_file = tempfile.NamedTemporaryFile(prefix='fasta_', suffix='.txt',
                                                delete=False)
        filename = temp_file.name

    with open(filename, 'w') as fle:
        for seq_id, seq in id_seqs.iteritems():
            fle.write('>' + str(seq_id) + '\n')
            fle.write(seq + '\n')

    return filename


def translate(seq, trans_table=CodonTable.unambiguous_dna_by_name["Standard"],
              min_prot_len=128):
    '''Translates supplied nucleotide sequence in all 6 reading frames.'''
    result = []

    seq = Seq(seq)

    for strand, nuc in [('+', seq), ('-', seq.reverse_complement())]:
        for frame in range(3):
            trans = \
                str(nuc[frame:-(len(nuc[frame:]) % 3)].translate(trans_table))
            trans_len = len(trans)
            aa_start = 0
            aa_end = 0

            while aa_start < trans_len:
                aa_end = trans.find("*", aa_start)
                if aa_end == -1:
                    aa_end = trans_len
                if aa_end - aa_start >= min_prot_len:
                    start = frame + aa_start * 3
                    end = frame + aa_end * 3

                    result.append((start, end, strand, frame,
                                   len(trans[aa_start:aa_end]),
                                   trans[aa_start:aa_end]))
                aa_start = aa_end + 1
    return result


def ambiguous_to_regex(seq):
    '''Converts sequence with ambiguous nucleotides to regex,
    e.g. ANT to A[ACGT]T.'''
    return ''.join([val
                    if val not in IUPACData.ambiguous_dna_values or
                    len(IUPACData.ambiguous_dna_values[val]) == 1
                    else '[' + IUPACData.ambiguous_dna_values[val] + ']'
                    for val in seq])


def apply_restriction(seq, restrict):
    '''Applies restriction site cleavage to forward and reverse strands.'''
    seq = _apply_restriction(seq, restrict)
    seq = _apply_restriction(get_rev_comp(seq), restrict)
    return get_rev_comp(seq)


def _apply_restriction(seq, restrict):
    '''Applies restriction site cleavage to sequence.'''
    match = re.search(restrict, seq)

    if match:
        seq = match.group(0)

    return seq


def _count_pattern(strings, pattern):
    '''Counts pattern in string of list of strings.'''
    if isinstance(strings, str) or isinstance(strings, unicode):
        return len(re.findall(pattern, strings))
    elif strings is None:
        return 0
    else:
        return [count_pattern(s, pattern) for s in strings]


def _scale(codon_usage):
    '''Scale codon usage values to add to 1.'''
    codon_usage = dict([(key, value / sum(codon_usage.values()))
                        for key, value in codon_usage.items()])

    return sorted(codon_usage.items(), key=operator.itemgetter(1),
                  reverse=True)


def _get_random_dna(length):
    '''Returns a random sequence of DNA of the supplied length.'''
    return ''.join(random.choice(['A', 'C', 'G', 'T']) for _ in range(length))


def _cleanup(drctry, pattern):
    '''Deletes files in directory matching pattern.'''
    for filename in os.listdir(drctry):
        if re.search(pattern, filename):
            os.remove(os.path.join(drctry, filename))


def _parse_uniprot_data(url, values):
    '''Parses Uniprot data.'''
    headers = None

    try:
        for line in urllib2.urlopen(url):
            tokens = line.strip().split('\t')

            if headers is None:
                headers = tokens
            else:
                resp = dict(zip(headers, tokens))
                entry = resp.pop('Entry')

                if 'Protein names' in resp:
                    regexp = re.compile(r'(?<=\()[^)]*(?=\))|^[^(][^()]*')
                    names = regexp.findall(resp.pop('Protein names'))
                    resp['Protein names'] = [nme.strip() for nme in names]

                values.update({entry: resp})
    except urllib2.HTTPError, err:
        print err


def _read_codon_usage_orgs_file(filename):
    '''Reads Codon Usage Database table of species file.'''
    codon_orgs = {}

    with open(filename, 'r') as textfile:
        next(textfile)

        for line in textfile:
            tokens = line.strip().split('\t')
            codon_orgs[tokens[0]] = tokens[1]

    return codon_orgs


def _expand_codon_usage_orgs(codon_orgs, max_errors=16):
    '''Expand Codon Usage Db table of species with children and synonyms.'''
    for tax_id in codon_orgs.values():
        for name in taxonomy.get_synonyms_by_id(tax_id):
            _add_codon_usage_org(codon_orgs, name, tax_id)

        errors = 0
        success = False

        while not success:
            try:
                for child in taxonomy.get_children_by_id(tax_id):
                    _add_codon_usage_org(codon_orgs, child['name'], tax_id)

                    for name in child['names']:
                        _add_codon_usage_org(codon_orgs, name, tax_id)

                success = True

            except ConnectionError, err:
                errors += 1

                if errors == max_errors:
                    raise err


def _add_codon_usage_org(codon_orgs, name, tax_id):
    '''Adds name to codon_orgs.'''
    if name not in codon_orgs:
        codon_orgs[name] = tax_id


def _write_codon_usage_orgs_file(codon_orgs, filepath):
    '''Writes Codon Usage Database table of species file.'''
    with open(filepath, 'w+') as fle:
        fle.write('Name\tId\n')

        for name, tax_id in codon_orgs.iteritems():
            fle.write(name + '\t' + tax_id + '\n')


def main():
    '''main method.'''
    get_codon_usage_organisms(expand=True)

if __name__ == '__main__':
    main()



'''
synbiochem (c) University of Manchester 2016

synbiochem is licensed under the MIT License.

To view a copy of this license, visit <http://opensource.org/licenses/MIT/>.

@author:  neilswainston
'''


def get_well(idx, rows=8, columns=12):
    '''Map idx to well'''
    if idx < 0 or idx >= rows * columns:
        raise ValueError('Index %idx out of range' % idx)

    return chr(ord('A') + (idx / rows)) + str(idx % rows + 1)
